/*
 * @(#) Jotm.java
 *
 * JOTM: Java Open Transaction Manager
 *
 * This project is developed inside the ObjectWeb Consortium,
 * http://www.objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jotm;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import javax.transaction.UserTransaction;

import org.objectweb.carol.util.configuration.CarolConfiguration;
import org.objectweb.carol.util.configuration.RMIConfigurationException;
import org.objectweb.transaction.jta.TMService;
import org.objectweb.transaction.jta.TransactionManager;

/**
 *  This  class represents an instance of JOTM.
 * 
 * @author jeff mesnil
*/
public class Jotm implements TMService {

    /**
     * Current instance
     */
    Current current = null;

    /** 
     * JNDI name of the TransactionFactory
     */
    private static final String TMFACTORY_NAME = "TMFactory";

    /**
     * TransactionFactory
     */
    private TransactionFactory tf = null;

    /**
     * is the TransactionFactory is local or not
     */
    private boolean local;

    /**
     * is the TransactionFacotory bound to JNDI or not
     */
    private boolean bound;

    /**
     * has binding of TransactionFactory fails
     */
    private boolean boundFailed = false;

    /**
     * Public constructor for Jotm.
     *
     * If Jotm is created with a <code>local</code> transaction factory which is
     * not <code>bound</code> to a registry, Jotm would be able to manage
     * transactions only inside the same JVM.
     * @param local <code>true</code> to create an instance of JOTM with a
     * local transaction factory, <code>false</code> else
     * @param bound <code>true</code> if the transaction factory is to be
     *  bound in a registry, <code>false</code> else (ignored if 
     * <code>local<code> is <code>false</code>)
     * 
     * @throws NamingException thrown if the transaction factory can't be bound
     * or looked up in a registry
     */
    public Jotm(boolean local, boolean bound) throws NamingException {
        if (TraceTm.jotm.isInfoEnabled()) {
            TraceTm.jotm.info(
                "JOTM started with a "
                    + (local ? "local" : "remote")
                    + " transaction factory"
                    + " which is "
                    + (bound ? "" : "not")
                    + " bound.");
        }
        this.local = local;
        this.bound = bound;
        // CAROL initialization
        TraceTm.jotm.info("CAROL initialization");
        try {
            CarolConfiguration.init();
        } catch (RMIConfigurationException e) {
            TraceTm.jotm.error("CAROL initialization failed", e);
        }
        if (local) {
            try {
                tf = new TransactionFactoryImpl();
            } catch (RemoteException e) {
                // should not happen: TransactionFactoryImpl throws 
                // RemoteException only because it extends Remote interface
                TraceTm.jotm.error(
                    "Instanciation of TransactionFactory failed",
                    e);
            }
            if (bound) {
                Context ictx;
                try {
                    ictx = new InitialContext();
                    ictx.rebind(TMFACTORY_NAME, tf);
                    TraceTm.jotm.debug(
                        "TransactionFactory bound with name " + TMFACTORY_NAME);
                } catch (NamingException e) {
                    TraceTm.jotm.error("TransactionFactory rebind failed", e);
                    boundFailed = true;
                    throw e;
                }
            }
        } else {
            Context ictx;
            try {
                ictx = new InitialContext();
                tf = (TransactionFactory) ictx.lookup(TMFACTORY_NAME);
            } catch (NamingException e) {
                TraceTm.jotm.error("TransactionFactory lookup failed", e);
                throw e;
            }
        }
        current = new Current(tf);
    }

    /**
     * @see org.objectweb.transaction.jta.TMService#getTransactionManager()
     */
    public TransactionManager getTransactionManager() {
        TraceTm.jotm.debug("TransactionManager=" + current);
        return (TransactionManager) current;
    }

    /**
     * @see org.objectweb.transaction.jta.TMService#getUserTransaction()
     */
    public UserTransaction getUserTransaction() {
        TraceTm.jotm.debug("UserTransaction=" + current);
        return (UserTransaction) current;
    }
    /**
     * @see org.objectweb.transaction.jta.TMService#stop()
     */
    public void stop() {
        TraceTm.jotm.info("stop JOTM");
        // remove current transaction if there's still one in ThreadLocal 
        current.forget();

        // unexport remote objects
        if (local) {
            if (bound && !boundFailed) {
                try {
                    InitialContext ictx = new InitialContext();
                    ictx.unbind(TMFACTORY_NAME);
                    TraceTm.jotm.debug("TransactionFactory unbound");
                } catch (Exception e) {
                    TraceTm.jotm.warn(
                        "an exception has prevented the TransactionFactory"
                            + " to be unbound",
                        e);
                }
            }
            if (tf != null) {
                try {
                    PortableRemoteObject.unexportObject(tf);
                    TraceTm.jotm.debug("TransactionFactory unexported");
                } catch (NoSuchObjectException e) {
                    TraceTm.jotm.warn(
                        "an exception has prevented the TransactionFactory"
                            + " to be unbound",
                        e);
                }
            }
        }
        tf = null;
    }
}
